import json
import pathlib
from abc import ABC, abstractmethod
from timeit import default_timer as timer

import imageio
import numpy as np


class Producer(ABC):
    """
    This is responsible for creating segmentation that will be later evaluated.
    It should be able to cache the results and add processing related tags.
    """

    def __init__(self, name, cache_root):
        # TODO add memory cache (lru based) and make cache root optional or replace it with some storage object
        self.cache_root = pathlib.Path(cache_root)
        self.name = name

    def load_segment(self, id):
        cache_path = (self.cache_root / str(id)).with_suffix(".tif")
        return imageio.imread(str(cache_path))

    def load_tag(self, id):
        cache_path = (self.cache_root / str(id)).with_suffix(".json")
        with open(str(cache_path), 'r') as f:
            return json.load(f)

    def __save_segment(self, id, segm):
        cache_path = (self.cache_root / str(id)).with_suffix(".tif")
        imageio.imsave(str(cache_path), segm)

    def __save_tag(self, id, tag):
        cache_path = (self.cache_root / str(id)).with_suffix(".json")
        with open(str(cache_path), 'w') as f:
            json.dump(tag, f)

    @abstractmethod
    def segmentation(self, image: np.ndarray) -> np.ndarray:
        """
        Args:
            image: image in uint8 (0-255)
        """
        # TODO for now assumes it is binary or 0-1 numeric
        pass

    def calculate(self, input_image: np.ndarray, input_tag: dict) -> (np.ndarray, dict):
        # TODO is possible and requested load results and tags from cache
        assert input_tag is not None
        if input_image.dtype != np.uint8: # TODO we need to be sure
            input_image = (input_image * 255).astype(np.uint8)
        assert input_image.dtype == np.uint8

        start_time = timer()
        seg = self.segmentation(input_image).astype(np.uint8)
        seg_tag = {}
        prediction_time = timer() - start_time
        seg_tag['run_time'] = prediction_time
        seg_tag['run_fps'] = round(1.0 / prediction_time, 2)
        seg_tag['producer_name'] = self.name
        seg_tag['producer_details'] = self.__repr__()

        if self.cache_root and "id" in input_tag:
            self.__save_tag(input_tag["id"], seg_tag)
            self.__save_segment(input_tag["id"], seg)

        return seg, seg_tag

    def __repr__(self):
        return f"{self.__class__} ({self.__dict__})"





import pathlib
import unittest

from tests.testbase import TestBase
from sep.loaders.images import ImagesLoader
import numpy.testing as nptest


class TestImagesLoader(TestBase):
    def test_loading(self):
        test_images_loader = ImagesLoader(self.test_dir("input/lights"))
        self.assertEqual(2, len(test_images_loader))
        self.assertEqual(['lights01', 'lights02'], test_images_loader.input_order)

        input_data_02_by_id = test_images_loader.load_image(1)
        input_data_02_by_name = test_images_loader.load_image('lights02')
        nptest.assert_equal(input_data_02_by_id, input_data_02_by_name)

        tag_02 = test_images_loader.load_tag('lights02')
        self.assertEqual("lights02", tag_02["id"])
        self.assertEqual("thenet", tag_02["source"])
        non_existing_tag10 = test_images_loader.load_tag('lights10')
        self.assertEqual("lights10", non_existing_tag10["id"])
        self.assertNotIn("source", non_existing_tag10)

        annotation_1 = test_images_loader.load_annotation(0)
        self.assertEqual(annotation_1.shape, input_data_02_by_id.shape[:2])
        self.assertEqual(255, annotation_1.max())

        tag_1 = test_images_loader.load_tag(0)
        self.assertEqual(0, tag_1["id"])  # TODO RETHINK default tags mirror exact call


if __name__ == '__main__':
    unittest.main()



from abc import ABC

from sep.loaders.loader import Loader


class MoviesLoader(Loader, ABC):
    """
    This one loads frames from the movies files and tags them so that they can be reorganizes into the files
    after evaluation. It also provides timestamp so that real time solution can be tested as well.
    """
    def __init__(self, data_root, framerate, clips_len, clips_count):
        super().__init__()


import os
import typing
from pathlib import Path

import pandas as pd
from tqdm import tqdm

from sep.assessors.detailer import Detailer
from sep.assessors.metricer import Metricer
from sep.loaders.loader import Loader
from sep.producers.producer import Producer


def evaluate(data_loader: Loader, producer: Producer, metricer: Metricer,
             detailer: typing.Optional[Detailer], output_evalpath,
             verbose=1):
    if verbose:
        print(f"Evaluation of {producer} on data from {data_loader}.")
        print(f"There are {len(data_loader)} images to evaluate on.")
    os.makedirs(output_evalpath, exist_ok=True)

    for i in tqdm(range(len(data_loader)), producer.name):
        image = data_loader.load_image(i)
        gt = data_loader.load_annotation(i)
        tag = data_loader.load_tag(i)

        if gt is None:
            print(data_loader.list_images()[i], "does not have annotation data!")

        segment, segment_tag = producer.calculate(image, tag)
        data_point_eval = metricer.evaluate_image(image, tag, segment, segment_tag, gt)
        if detailer is not None:  # TODO detailer vs metricer relation has to be rethought
            detailer.save_image_evaluation(data_point_eval)

    return metricer.report_overall()


def compare(data_loader: Loader, producers: typing.List[Producer], metricer, detailer, output_evalpath):
    print(f"Comparison of {len(producers)} producers on data from {data_loader}.")

    reports = []
    for producer in producers:
        producer_eval_path = Path(output_evalpath) / producer.name
        producer_report = evaluate(data_loader, producer, metricer, detailer, producer_eval_path,
                                   verbose=0)
        producer_report.insert(loc=0, column='producer', value=producer.name)

        reports.append(producer_report)
    return pd.concat(reports).groupby(['producer', 'region']).mean()



from abc import ABC, abstractmethod


class Loader(ABC):
    def __init__(self):
        self.annotator = None
        pass

    @abstractmethod
    def list_images(self):
        pass

    @abstractmethod
    def load_image(self, name_or_num):
        pass

    @abstractmethod
    def load_tag(self, name_or_num):
        pass

    @abstractmethod
    def load_annotation(self, name_or_num):
        pass

    def save_annotation(self, name_or_num, new_annotation, keep_history=False):
        pass

    def save_tag(self, name_or_num, new_tag):
        pass

    def __len__(self):
        return len(self.list_images())



import json
import os
import pathlib
from glob import glob

import imageio

from sep.loaders.loader import Loader


class ImagesLoader(Loader):
    """
    Look through entire file structure in the data_root path and collect all the images.
    """

    def __init__(self, data_root, image_extensions=None, annotation_suffix="_gt", annotation_extension=".png"):
        super().__init__()
        self.data_root = data_root
        self.image_extensions = image_extensions or [".tif", ".png", ".jpg"]
        all_files = [pathlib.Path(p) for p in sorted(glob(os.path.join(data_root, "**", "*.*"), recursive=True))]

        input_images_paths = [f for f in all_files
                              if f.suffix in self.image_extensions and not f.stem.endswith(annotation_suffix)]

        self.input_images = {self.path_to_id(p): p for p in input_images_paths}
        self.input_order = sorted(self.input_images.keys())
        self.annotation_images = {}
        self.json_tags = {}
        for input_path in input_images_paths:
            annotation_path = input_path.with_name(input_path.stem + annotation_suffix + annotation_extension)
            if os.path.isfile(annotation_path):
                self.annotation_images[self.path_to_id(input_path)] = annotation_path

            json_path = input_path.with_suffix(".json")
            if os.path.isfile(json_path):
                self.json_tags[self.path_to_id(input_path)] = json_path

    def path_to_id(self, path):
        return path.stem  # TODO this may not be unique

    def list_images(self):
        return list(self.input_order)

    def list_images_paths(self):
        return [self.input_images[p] for p in self.input_order]

    def __get_file_path(self, path_set, name_or_num):
        if isinstance(name_or_num, int):
            name_or_num = self.input_order[name_or_num]
        if isinstance(name_or_num, str):
            return path_set.get(name_or_num, None)
        else:
            raise NotImplemented(type(name_or_num))

    def load_image(self, name_or_num):
        path_to_file = self.__get_file_path(self.input_images, name_or_num)
        return imageio.imread(path_to_file)

    def load_tag(self, name_or_num):
        path_to_file = self.__get_file_path(self.json_tags, name_or_num)
        if path_to_file is None:
            return {"id": name_or_num}
        else:
            with open(str(path_to_file), 'r') as f:
                tag = json.load(f)
            assert "id" in tag
            return tag

    def load_annotation(self, name_or_num):
        path_to_file = self.__get_file_path(self.annotation_images, name_or_num)
        if path_to_file is None:
            return None
        return imageio.imread(path_to_file)

    def __str__(self):
        return f"ImageLoader for: {self.data_root}"



from abc import abstractmethod, ABC

import numpy as np

class Region(ABC):
    """
    This class generate the transformations of the segmentation and ground truth so that they can be evaluated
    in the same manner as the entire image. E.g. this can be used to generate metrics on only edges of the ground
    truth mask.
    """
    def __init__(self, name):
        self.name = name

    @abstractmethod
    def regionize(self, ground_truth: np.ndarray, mask: np.ndarray) -> np.ndarray:
        return mask

    def __str__(self):
        return self.name


class EntireRegion(Region):
    def __init__(self):
        super().__init__("Entire image")

    def regionize(self, ground_truth, mask) -> np.ndarray:
        return mask

